/*
 * $LastChangedBy$
 * $LastChangedDate$
 * $LastChangedRevision$
 * $HeadURL$
 *
 * Copyright 2014 BlueConic Inc./BlueConic B.V. All rights reserved.
 */
extension JSON {
    var method: Method? {
        get {
            switch self.type {
            case .string:
                return  Method(rawValue:self.object as? String ?? "")
            default:
                return nil
            }
        }
        set {
            self.object = NSString(string:newValue!.rawValue)
        }
    }
}

enum Method: String {
    case Add = "add",
    Sum = "merge",
    Set = "set"
    var description: String {
        return self.rawValue
    }
}

import UIKit
import Foundation
private func < <T: Comparable>(lhs: T?, rhs: T?) -> Bool {
    switch (lhs, rhs) {
    case let (l?, r?):
        return l < r
    case (nil, _?):
        return true
    default:
        return false
    }
}

public class RuleService: EventSerivceBase {
    private let LOG_TAG = "BC_RULESERVICE"
    private var _profileChanges: [ProfileChange] = []
    private var _eventManager: BlueConicEventManager!
    private var _clickRuleMap = [UIControl: [JSON]]()
    
    public struct ProfileChange {
        var property: String
        var values: [String]
        var method: Method
        var description: String {
            return "Property: \(property), values \(values), method: \(method.description)"
        }
    }
    
    /**
     Inits everything needed for the engagementservice
     - parameter listenerId: The id of the listener
     */
    @objc public convenience init(listenerId: String) {
        self.init(BlueConicClient.getInstance(nil), listenerId: listenerId)
    }
    
    /**
     Convenience class to apply rules. This is then transformed into a json array object my mapping each arraymember into a JSON object.
     - parameter rules: An array of Any
     */
    @objc public func applyRules(_ rules: [Any]) {
        self.applyRules(rules.map({ JSON($0) }))
    }
    
    /*
     Loops over all available rules and applies them when possible
     - parameter rules: An array of rules in a JSON object
     */
    public func applyRules(_ rules: [JSON]) {
        for rule in rules {
            self.BCLog("Handling \(rule)")
            
            if let ruleType: String = rule[PluginConstants.Listener.TAG_RULE_TYPE].string {
                switch ruleType {
                case PluginConstants.Listener.RULETYPE_SCORE_CONTENT, "":
                    self.handleContentRule(rule)
                case PluginConstants.Listener.RULETYPE_SCORE_CLICK:
                    if let clickSelector = ListenerUtil.getSelector(rule[PluginConstants.Listener.TAG_CLICKAREA]) {
                        self.BCLog("Adding target to \(clickSelector)")
                        // We need to use var to addTarget
                        if let element = self._client.getView(clickSelector) as? UIControl {
                            // map objects with their rule, since we can't set parameters with addTarget
                            element.addTarget(self, action: #selector(RuleService.touchAction(_:)), for: UIControl.Event.touchDown)
                            self.addToClickRuleMap(element, rule: rule)
                            self.registerEvent(rule, className: NSStringFromClass(ClickEvent.self))
                        } else {
                            self.registerEvent(rule, className: NSStringFromClass(ClickEvent.self))
                            self.BCLog("Element from \(clickSelector) not clickable, registering event for custom ClickEvents")
                        }
                    } else {
                        self.BCLog("Found click rule without clickarea selector: \(rule)")
                    }
                case PluginConstants.Listener.RULETYPE_SCORE_FORMSUBMIT:
                    if let clickSelector = ListenerUtil.getSelector(rule[PluginConstants.Listener.TAG_FORM]) {
                        // We need to use var to addTarget
                        if let element = self._client.getView(clickSelector) as? UIControl {
                            // map objects with their rule, since we can't set parameters with addTarget
                            element.addTarget(self, action: #selector(RuleService.touchAction(_:)), for: UIControl.Event.touchDown)
                            self.addToClickRuleMap(element, rule: rule)
                            self.registerEvent(rule, className: NSStringFromClass(FormSubmitEvent.self))
                        } else {
                            self.BCLog("Element from \(clickSelector) not clickable")
                        }
                    } else {
                        self.BCLog("Found form submit rule without form selector: \(rule)")
                    }
                case PluginConstants.Listener.RULETYPE_SCORE_URL:
                    var content = ""
                    if let urlRule = rule[PluginConstants.Listener.TAG_URL].string {
                        // We can't reliably check referrers.
                        if urlRule == "url" || urlRule == "or" || urlRule == "urlreferrer" {
                            let words = ListenerUtil.getWords(rule)
                            if self._client.getScreenName() != "" {
                                content += self._client.getScreenName()
                            } else {
                                BCLog("Unable to get the current screen name with the url rule: \(rule)")
                            }
                            
                            if ListenerUtil.contentContainsWord(rule, content:[content], interests: words) {
                                self.addProfilechange(rule)
                            }
                        }
                    }
                case PluginConstants.Listener.RULETYPE_SCORE_EVENT, PluginConstants.Listener.RULETYPE_SCORE_SOCIAL_EVENT:
                    self.registerEvent(rule, className: NSStringFromClass(AdvancedEvent.self))
                default:
                    self.BCLog("Invalid or not supported rule detected", flag: Constants.Flag.WARN)
                }
            }
        }
        
        subscribeListeners();
    }
    
    private func getValues(_ rule: JSON) -> [String]? {
        var values: [String] = []
        
        if let value = rule[PluginConstants.Listener.TAG_VALUES].string, value == PluginConstants.Listener.TAG_DATETIME {
            //Set datetime values
            values = [ListenerUtil.getCurrentTime()]
        } else if let value = rule[PluginConstants.Listener.TAG_VALUES].arrayObject as? [String] {
            // set array of values
            values = value
        } else if let value = rule[PluginConstants.Listener.TAG_VALUES][PluginConstants.Listener.TAG_SELECTED_OPTION].string {
            if value == "regexp" {
                // need getlocation
                let location = self._client.getScreenName()
                let nsString = location as NSString
                let regexString = rule[PluginConstants.Listener.TAG_VALUES]["regexp"].stringValue
                if let regex  = try? NSRegularExpression(pattern: regexString, options: []) {
                    let results: [NSTextCheckingResult] = regex.matches(in: location, options: [], range: NSMakeRange(0, nsString.length))
                    let finalResult = results.map {nsString.substring(with: $0.range)}
                    
                    values = finalResult.filter({$0 != ""})
                } else {
                    self.BCLog("No valid reg. exp: '\(regexString)'", flag: Constants.Flag.WARN)
                }
                
            } else if value == PluginConstants.Listener.TAG_SELECTOR {
                values = ListenerUtil.getContent(rule[PluginConstants.Listener.TAG_VALUES][PluginConstants.Listener.TAG_SELECTOR].stringValue)
            } else if value == "parameter" {
                // not suited to mobile
            }
        }
        
        return values.count > 0 ? values : nil
    }
    
    private func addProfilechange(_ rule: JSON) {
        if let profileProperty: String = rule[PluginConstants.Listener.TAG_PROFILE_PROPERTY][0][PluginConstants.Listener.TAG_PROFILE_PROPERTY].string {
            if let addSet = rule[PluginConstants.Listener.TAG_ADD_SET].method, let values = getValues(rule) {
                self._profileChanges.append(ProfileChange(property: profileProperty, values: values, method: addSet))
            }
        }
    }
    
    /**
     The changes object only exists if there have been any changes, convenience function to check this.
     
     - returns:  bool that denotes a change has occured
     */
    @objc public func isChanged() -> Bool {
        return !self._profileChanges.isEmpty
    }
    
    private func getFirstNonEmptyValue(_ values: [String]?) -> String {
        if let values = values {
            for value: String in values where value != "" {
                return value.trimmingCharacters(in: CharacterSet.whitespaces)
            }
        }
        // when there are no values, start counting at zero
        return "0"
    }
    
    /**
     Save points to BlueConic by setting the changes object to a property.
     */
    @objc public func save() {
        if self.isChanged() {
            for profileChange: ProfileChange in self._profileChanges {
                self.saveChange(profileChange)
            }
        }
        self._profileChanges = []
        
    }
    
    /**
     Handles an event (eg click event, advanced event), which leads to points for keywords
     - parameter event: Event object
     */
    public override func handleEvent(_ event: Event) {
        self.BCLog("Handling event \(event)")
        if let rules = getRules(event) {
            for rule in rules {
                if let advancedEvent = event as? AdvancedEvent {
                    self.handleAdvancedEvent(rule, advancedEvent: advancedEvent)
                } else if let updateContentEvent = event as? UpdateContentEvent {
                    self.handleUpdateContentEvent(rule, updateContentEvent: updateContentEvent)
                } else if let clickEvent = event as? ClickEvent {
                    self.handleTouchEvent(rule, clickEvent: clickEvent )
                } else if let formSubmitEvent = event as? FormSubmitEvent {
                    self.handleTouchEvent(rule, clickEvent: formSubmitEvent)
                }
                self.save()
            }
            
        }
    }
    
    private func saveChange(_ change: ProfileChange) {
        self.BCLog("Profile Change: \(change.description)")
        if change.method == Method.Set {
            // Set
            self._client.setProfileValues(change.property, values: change.values)
        } else if change.method == Method.Sum {
            // sum
            let currentValue = getFirstNonEmptyValue(self._client.getProfileValues(change.property))
            if let summedValue = sum(currentValue, newValues: change.values) {
                self.BCLog("Summed value: \(summedValue)")
                self._client.setProfileValue(change.property, value: summedValue)
            } else {
                // do nothing if we can't sum
            }
        } else {
            self._client.addProfileValues(change.property, values: change.values)
        }
    }
    
    private func sum(_ currentValue: String, newValues: [String]) -> String? {
        self.BCLog("Merging currentValue \(currentValue) with \(newValues)")
        if let currentDouble: Double = Double(convertToFormattable(currentValue)) {
            let valueMap: [Double?] = newValues.map { return Double(self.convertToFormattable($0)) }
            return round(valueMap.filter {$0 != nil}.reduce(currentDouble) {$0 + $1!}).description
        }
        return nil
    }
    
    private func BCLog(_ message: String, flag: String = Constants.Flag.INFO) {
        BlueConicConfiguration.log(message, tag: self.LOG_TAG, flag: flag)
    }
    
    /**
     Handles an touch event rule as set in blueconic. Similar to handleEventRule, handleContentRule, etc. Unlike other rule handlers this is called when a touch action is detected
     
     - parameter touchRule:  A Json object describing the rule as configured in blueconic
     - parameter clickEvent:  A click event
     */
    private func handleTouchEvent(_ touchRule: JSON, clickEvent: ClickEvent) {
        var words: [String] = ListenerUtil.getWords(touchRule)
        
        if let clickArea = ListenerUtil.getSelector(touchRule[PluginConstants.Listener.TAG_CLICKAREA]), clickArea == clickEvent.selector,
            let selector = ListenerUtil.getSelector(touchRule[PluginConstants.Listener.TAG_CONTENT_AREA]) {
            if words[0] == PluginConstants.Listener.PRE_ANY && (selector == PluginConstants.Listener.PRE_ANY || selector == PluginConstants.Listener.ANY) {
                addProfilechange(touchRule)
            } else if ListenerUtil.contentContainsWord(touchRule, content:ListenerUtil.getContent(selector), interests: words) {
                addProfilechange(touchRule)
            }
        }
    }
    
    /**
     Handles the content rule when the content is updated with the update content event
     - parameter rule: JSON object of the content rule as defined in the listener
     - parameter updateContentEvent: The update content event
     */
    private func handleUpdateContentEvent(_ rule: JSON, updateContentEvent: UpdateContentEvent) {
        self.handleContentRule(rule, content: updateContentEvent.content)
    }
    
    
    /**
     Handles the content rule when the content is updated by a Content event, use the function above when updated based on UpdateContentEvent
     - parameter rule: JSON object of the content rule as defined in the listener
     */
    private func handleContentRule(_ rule: JSON) {
        self.handleContentRule(rule, content: nil)
    }
    
    /**
     Handles the content rule when the content is updated
     - parameter rule: JSON object of the content rule as defined in the listener
     - parameter content: A list of content items, this is the content of an UpdateContentEvent, will be nil when it is handled by the default content rule.
     */
    private func handleContentRule(_ rule: JSON, content: [String]?) {
        if let selector = ListenerUtil.getSelector(rule[PluginConstants.Listener.TAG_CONTENT_AREA]) {
            let words = ListenerUtil.getWords(rule)
            let ruleContent: [String] = content ?? ListenerUtil.getContent(selector)
            if ListenerUtil.contentContainsWord(rule, content: ruleContent, interests: words) {
                self.addProfilechange(rule)
            }
            self.registerEvent(rule, className: NSStringFromClass(UpdateContentEvent.self))
        } else {
            self.BCLog("Found content rule without selector: \(rule)")
        }
    }
    
    private func handleAdvancedEvent(_ rule: JSON, advancedEvent: AdvancedEvent) {
        if let event = rule[PluginConstants.Listener.TAG_EVENT].string, event == advancedEvent.eventName {
            var contextContent: [String] = []
            if let eventContext = advancedEvent.context {
                if let contextPosition = rule[PluginConstants.Listener.TAG_CONTEXT_POSITION].string,
                    let contextPositionNr = Int(contextPosition) {
                    // When it has been provided, only match with the value at the given context position
                    contextContent = [eventContext[contextPositionNr - 1]]
                } else {
                    contextContent = eventContext
                }
            }
            let words = ListenerUtil.getWords(rule)
            if ListenerUtil.contentContainsWord(rule, content: contextContent, interests: words) {
                self.addProfilechange(rule)
            }
        }
    }
    
    /**
     Respond to touchAction
     - parameter sender: the UI element that triggered the event
     */
    @objc public func touchAction(_ sender: UIControl) {
        if let touchRules = self._clickRuleMap[sender] {
            for touchRule in touchRules {
                if let selector = touchRule[PluginConstants.Listener.TAG_CLICKAREA][PluginConstants.Listener.TAG_SELECTOR].string {
                    BlueConicEventFactory.getInstance().publish(ClickEvent(selector:selector))
                    BlueConicEventFactory.getInstance().publish(FormSubmitEvent(selector:selector))
                    break
                } else if let selector = touchRule[PluginConstants.Listener.TAG_FORM][PluginConstants.Listener.TAG_SELECTOR].string {
                    BlueConicEventFactory.getInstance().publish(FormSubmitEvent(selector:selector))
                    BlueConicEventFactory.getInstance().publish(ClickEvent(selector:selector))
                    break
                }
            }
        }
    }
    
    private func addToClickRuleMap(_ element: UIControl, rule: JSON) {
        if self._clickRuleMap[element] == nil {
            self._clickRuleMap[element] = [rule]
        } else {
            var rules: [JSON] = self._clickRuleMap[element]!
            rules.append(rule)
            self._clickRuleMap[element] = rules
        }
    }
    
    private func convertToFormattable(_ value: String) -> String {
        let hasDot = value.range(of: ".", options: String.CompareOptions.backwards)
        let hasComma = value.range(of: ",", options: String.CompareOptions.backwards)
        
        if hasDot != nil && hasComma != nil {
            var v = ""
            if hasDot?.lowerBound < hasComma?.lowerBound {
                v = value.replacingOccurrences(of: ".", with: "")
            } else {
                v = value.replacingOccurrences(of: ",", with: "")
            }
            return v.replacingOccurrences(of: ",", with: ".")
        } else if hasDot != nil || hasComma != nil {
            let divider = hasDot != nil ? "." : ","
            
            var valueParts = value.components(separatedBy: divider)
            if valueParts.count != 2 {
                return value.replacingOccurrences(of: divider, with: "")
            } else if valueParts[0].count <= 3 && valueParts[1].count == 3 {
                return value.replacingOccurrences(of: divider, with: "")
            } else if valueParts[1].count != 3 {
                return value.replacingOccurrences(of: divider, with: ".")
            }
        }
        return value
    }
}
